''''''''''''''''''''''''''''''''''''''''''''''''''
' Copyright www.acecomputing.biz
' Feel free to use and modify this code for private use.
' Not licensed for resale
' This code has been inspired by code from Paras Chopra and Aldo Vagras
'''''''''''''''''''''''''''''''''''''''''''''''''''
Imports System.Collections

Public Class PDFDocument
    Public ReadOnly _copyrightAceComputing As String = "www.acecomputing.biz"
#If Debug Then
    Private eol = vbCrLf
#Else
    Private eol = vbLf
#End If
    Dim _Author As String = ""
    Dim _Creator As String = ""
    Dim _Producer As String = "www.acecomputing.biz"
    Dim _FileName As String = ""
    Dim _Title As String = ""
    Dim _keywords As String = ""
    Dim _subject As String = ""
    Dim _FontAlias As Hashtable      ' One entry per font
    Dim _PageNumber As Integer

    Dim Position As Integer
    Dim _PDFLocation(5000) As Integer       ' Positions of all the PDF objects
    Dim pageObj(5000) As Integer            ' Page objects
    Dim obj As Integer                      ' PDF objects

    Dim _rootObj As Integer         ' RootObject is the object after properties
    Dim _TopPagesObj As Integer     ' Top page comes after rootobject
    Dim _EncodingObj As Integer     ' Object for Encoding type
    Dim _PropObj As Integer
    Dim cache As String

    Dim _controls As ArrayList
    Dim _PageHeight As Integer = 792
    Dim _Pagewidth As Integer = 612
    Dim _drawableWidth As Integer
    Dim _TopMargin As Integer = 54  ' 3/4 inch, An adobe document has another 1/4 inch built in margin
    Dim _LeftMargin As Integer = 72 ' 1 inch, An adobe document has another 1/4 inch built in margin

    Public Sub New()
        _controls = New ArrayList()
        _FontAlias = New Hashtable()
        obj = 0
        Position = 0
        cache = ""
    End Sub

    Public ReadOnly Property PageWidth() As Integer
        Get
            Return _Pagewidth / 72
        End Get
    End Property

    Public Sub AddControl(ByVal control As iDrawableObject)
        Dim ta As TextArea
        If TypeOf control Is TextArea Then
            ta = control.GetCopy
            _controls.Add(ta)
        Else
            _controls.Add(control)
        End If

    End Sub

    Public Sub OutputToFile(Optional ByVal filename As String = "")
        If filename <> "" Then
            _FileName = filename
        End If
        If FileExists(_FileName) Then
            Kill(_FileName)
        End If
        Call WriteStart()
        Call WriteHead()
        Call WritePage()
        Call endPDF()
    End Sub

    Private Function WritePage()
        Dim beginstream As String, fonts As String, fontRef As String, key As String
        Dim pdfo As PDFObject, fonto As FontObj, resources As String
        Dim contents As String, stream As String, StartY As Integer, width As Integer
        Dim control As iDrawableObject, dummy As String, page As PageBreak
        Dim myEnumerator As System.Collections.IEnumerator = _
           _controls.GetEnumerator()
        Dim PageFonts As PDFObject
        Dim TopStart As Integer

        PageFonts = New PDFObject()
        fonts = " /Font << "
        StartY = _PageHeight - _TopMargin
        TopStart = StartY
        width = _Pagewidth - 2 * _LeftMargin
        While myEnumerator.MoveNext()
            control = myEnumerator.Current
            dummy = control.toString            ' Debug statement
            If control.StartPDFH = 0 Then
                control.StartPDFH = Me._LeftMargin
            End If
            pdfo = control.draw(StartY, width, _FontAlias, _PageNumber, TopStart)

            If pdfo.count > 1 Then
                stream += pdfo.getStream()
                StartPage(contents, resources, stream, fonts)
                stream = ""
                PageFonts = New PDFObject()
                fonts = " /Font << "
            End If
            stream += pdfo.getStream()

            Call WriteNewFonts()
            For Each fontRef In pdfo._fonts
                fonto = _FontAlias.Item(fontRef)
                If PageFonts.FontExists(fonto.FontObj) = False Then
                    PageFonts._fonts.Add(fonto.FontObj)
                    fonts += "/F" & fontRef & fonto.FontObj & " 0 R "
                End If
            Next
        End While

        If Len(stream) Then
            page = New PageBreak()
            page.StartPDFH = _LeftMargin
            pdfo = page.Draw(StartY, width, _FontAlias, _PageNumber, TopStart)
            stream += pdfo.getStream()
            StartPage(contents, resources, stream, fonts)
        End If

    End Function

    Private Sub StartPage(ByVal contents As String, ByVal resources As String, _
        ByVal stream As String, ByVal fonts As String)
        fonts += ">>"
        resources += fonts + vbCrLf
        resources += "/Procset [/PDF /Text]"

        obj = obj + 1
        contents += obj & " 0 R"
        _PDFLocation(obj) = Position
        writepdf(obj & " 0 obj")
        writepdf("<< /Length " & Len(stream) & " >>")
        writepdf("stream")
        writepdf(stream)
        writepdf("endstream")
        writepdf("endobj")

        obj = obj + 1
        _PDFLocation(obj) = Position
        ' pageNo = pageNo + 1
        pageObj(Me._PageNumber) = obj
        writepdf(obj & " 0 obj")
        writepdf("<<")
        writepdf("/Type /Page")
        writepdf("/Parent " & _TopPagesObj & " 0 R")
        writepdf("/Resources << " & resources & " >> ")
        writepdf("/Contents " & contents)
        'writepdf("/Rotate 0")
        writepdf(">>")
        writepdf("endobj")
    End Sub

    Private Sub WriteNewFonts()
        Dim i As Integer, fonts As String, key As String, fonto As FontObj
        Dim fontName As String, fontNumber As Integer, sobj As Integer
        sobj = obj
        For i = 1 To _FontAlias.Count
            key = Trim(Str(i))
            fonto = _FontAlias.Item(key)
            If fonto.FontObj Is Nothing Then
                obj = obj + 1
                fonto.FontObj = Str(obj)
                _PDFLocation(obj) = Position
                writepdf(obj & " 0 obj")
                writepdf("<<")
                writepdf("/Type /Font")
                writepdf("/Subtype /Type1")       ' Adobe Type 1
                writepdf("/Name /F" & fonto.FontRef)
                writepdf("/BaseEncoding /WinAnsiEncoding")
                writepdf("/BaseFont /" & fonto.FontName)
                writepdf(">>")
                writepdf("endobj")
            End If
        Next
    End Sub

    Private Sub WriteHead()
        WriteProperties()
        obj = obj + 1
        _rootObj = obj                  ' The root object will be written at the end
        obj = obj + 1
        _TopPagesObj = obj              ' The Pages object will be written at the end
        obj = obj + 1
    End Sub

    Private Sub writepdf(ByRef stre As String, Optional ByRef flush As Boolean = False)
        ' On Error Resume Next
        Dim i As Integer
        Position = Position + Len(stre) ' Position for the next object
        cache = cache & stre & eol
        If Len(cache) > 32000 Or flush Then
            i = FreeFile()
            FileOpen(i, _FileName, OpenMode.Append)
            Print(i, cache)
            FileClose(i)
            cache = ""
        End If
    End Sub

    Private Sub WriteStart()
        writepdf("%PDF-1.2")  ' Acrobat version 3.0
        writepdf("%")
    End Sub

    Sub endPDF()
        Dim ty As String
        Dim i As Short
        Dim xreF As Integer

        _PDFLocation(_rootObj) = Position
        writepdf(_rootObj & " 0 obj")
        writepdf("<<")
        writepdf("/Type /Catalog")
        writepdf("/Pages " & _TopPagesObj & " 0 R")
        writepdf(">>")
        writepdf("endobj")

        _PDFLocation(_TopPagesObj) = Position
        writepdf(_TopPagesObj & " 0 obj")
        writepdf("<<")
        writepdf("/Type /Pages")
        writepdf("/Count " & Me._PageNumber)
        writepdf("/MediaBox [ 0 0 " & _Pagewidth & " " & _PageHeight & " ]")
        ty = ("/Kids [ ")
        For i = 1 To Me._PageNumber
            ty = ty & pageObj(i) & " 0 R "
        Next i
        ty = ty & "]"
        writepdf(ty)
        writepdf(">>")
        writepdf("endobj")

        ' Xref
        xreF = Position
        writepdf("0 " & obj + 1)
        writepdf("0000000000 65535 f ")
        For i = 1 To obj
            'writepdf(VB6.Format(_PDFLocation(i), "0000000000") & " 00000 n ")
            writepdf(String.Format("{0:D10} 00000 n", _PDFLocation(i)))
        Next i

        ' Trailer
        writepdf("trailer")
        writepdf("<<")
        writepdf("/Size " & obj + 1)
        writepdf("/Root " & _rootObj & " 0 R")
        writepdf("/Info " & _PropObj & " 0 R")
        writepdf(">>")
        writepdf("startxref")
        writepdf(CStr(xreF))
        writepdf("%%EOF", True)
    End Sub

    Private Sub WriteProperties()
        Dim CreationDate As String

        'CreationDate = "D:" & VB6.Format(Now, "YYYYMMDDHHNNSS")
        CreationDate = "D:" & Format$(Now, "YYYYMMDDHHNNSS")
        obj = obj + 1
        _PDFLocation(obj) = Position
        _PropObj = obj

        writepdf(obj & " 0 obj")
        writepdf("<<")
        writepdf("/Author (" & _Author & ")")
        writepdf("/CreationDate (" & CreationDate & ")")
        writepdf("/Creator (" & _Creator & ")")
        writepdf("/Producer (" & _Producer & ")")
        writepdf("/Title (" & _Title & ")")
        writepdf("/Subject (" & _subject & ")")
        writepdf("/Keywords (" & _keywords & ")")
        writepdf(">>")
        writepdf("endobj")
    End Sub

    Public Function FileExists(ByVal filename As String) As Boolean
        On Error Resume Next
        FileExists = FileLen(filename) > 0
        Err.Clear()
    End Function

    Public Property Author() As String
        Get
            Return _Author
        End Get
        Set(ByVal PDFAuthor As String)
            _Author = PDFAuthor
        End Set
    End Property

    Public Property Creator() As String
        Get
            Return _Creator
        End Get
        Set(ByVal PDFCreator As String)
            _Creator = PDFCreator
        End Set
    End Property

    Public ReadOnly Property Producer() As String
        Get
            Return _Producer
        End Get
    End Property

    Public Property OutputFileName() As String
        Get
            Return _FileName
        End Get
        Set(ByVal FileName As String)
            _FileName = FileName
        End Set
    End Property
      
End Class
