''''''''''''''''''''''''''''''''''''''''''''''''''
' Copyright www.acecomputing.biz
' Feel free to use and modify this code for private use.
' Not licensed for resale
'''''''''''''''''''''''''''''''''''''''''''''''''''
Public Enum Borders
    thick
    thin
    none
End Enum

Public Class Table
    Implements iDrawableObject
    Public ReadOnly _copyrightAceComputing As String = "www.acecomputing.biz"
    Private _border As Borders = Border.thick
    Private _rows As ArrayList
    Private _height As Integer
    Private _cellPadding As Integer = 4
    Private _ColumnWidth As Integer               ' PDF measurement
    Private _cellCount As Integer
    Private _ActualHeight As Integer
    Private _StartH As Integer
    Private _StartV As Integer

    Public Sub New()
        _rows = New ArrayList()
    End Sub

    Function GetCopy() As iDrawableObject Implements iDrawableObject.GetCopy

    End Function

    Public Property StartPDFH() As Integer Implements iDrawableObject.StartPDFH
        Get
            Return _StartH
        End Get
        Set(ByVal MyStartInPDFUnits As Integer)
            _StartH = MyStartInPDFUnits
        End Set
    End Property

    Public Function Draw(ByRef StartV As Integer, ByVal width As Integer, ByRef FontAlias As Hashtable, _
        ByRef PageNum As Integer, ByVal TopMargin As Integer) As PDFObject Implements iDrawableObject.draw
        Dim pdfObj As PDFObject, row As Row, count As Integer, TotalCols As Integer
        Dim Stream As String            ' Text Stream
        Dim GStream As String           ' graphics stream
        Dim cell As Cell
        Dim RightH As Integer
        Dim V As Integer, H As Integer
        Dim RowStartV As Integer
        Dim cols As Integer
        Dim c As Integer
        Dim accumColumn As Integer
        Dim RowStarty As Integer

        pdfObj = New PDFObject()

        Call CalculateTable(width)
        ' Save start point
        _StartV = StartV
        If _border <> Borders.none Then

            Stream = "0.0 G " + vbCr       ' Black color
            If _border = Borders.thick Then
                Stream = "2 w " + vbCr       ' Line width
            Else
                Stream = "1 w " + vbCr       ' Line width
            End If

            RightH = _StartH + width

            ' Top level line of the table
            Stream += line(_StartH, StartV, RightH, StartV)

            For Each row In _rows

                ' Print first vertical bar for each cell
                V = StartV - row.HeightinPDFUnits
                Stream += line(_StartH, StartV, _StartH, V)

                ' Print right vertical bar for each cell
                cols = 0
                c = 1
                accumColumn = 0
                For Each cell In row.cells
                    cols = cols + cell.ColumnSpan
                    If c = row.cells.Count Then
                        H = RightH
                    Else
                        If cell.WidthInPercent = 0 Then
                            H = _StartH + cols * _ColumnWidth
                        Else
                            accumColumn += cell.WidthInPercent * width
                            H = _StartH + accumColumn
                        End If
                    End If
                    V = StartV - row.HeightinPDFUnits
                    Stream += line(H, StartV, H, V)
                    c = c + 1
                Next

                ' Print row divider
                StartV = StartV - row.HeightinPDFUnits
                Stream += line(_StartH, StartV, RightH, StartV)
            Next
        End If

        ' Print text in cells
        V = _StartV
        For Each row In _rows
            H = _StartH
            For Each cell In row.cells
                cell.StartPDFH = H + _cellPadding
                cell.StartPDFV = V
                pdfObj = cell.Draw(FontAlias, 1, TopMargin)
                Stream += pdfObj.getStream
                If cell.WidthInPercent = 0 Then
                    H = H + cell.ColumnSpan * _ColumnWidth
                Else
                    H = H + width * cell.WidthInPercent
                End If
            Next
            V = V - row.HeightinPDFUnits
        Next
        pdfObj.addStream(Stream)
        Draw = pdfObj
    End Function

    Friend Sub CalculateTable(ByVal width As Integer)
        Dim row As Row

        _ActualHeight = 0
        ' Calculate table width

        If width = 0 Then
            Throw New Exception("Zero Width table not supported.")
        End If

        ' Check to see that we have a column count
        If _rows.Count < 1 Then
            Throw (New Exception("No Rows to draw."))
        End If

        _cellCount = CalculateCellCount()

        ' Column width when all columns have the same width
        _ColumnWidth = (width - 2 * _border) / _cellCount

        ' Calculate
        For Each row In _rows
            row.CalculateHeight(_ColumnWidth, _cellPadding)
            'row.HeightinPDFUnits = row.HeightinPDFUnits
            ' row.Height = row.Height + TableBorderToPDF(row.Border)
            _ActualHeight = _ActualHeight + row.HeightinPDFUnits + 2 * _border
        Next
        _ActualHeight = _ActualHeight + 2 * _border

    End Sub

    Public Sub setColumnWidth(ByVal Width() As Integer)
        ' This method sets the width of the table columns
        ' Columns are from index 1 to the upper bound of width(). With(0) is not used.
        ' Each entry in the input array becomes a percentage of the sum of all entries in the input array
        Dim row As Row, cell As Cell, totalWidth As Integer, I As Integer
        Dim cols As Integer

        If _rows.Count < 1 Then
            Throw (New Exception("No rows."))
        End If

        _cellCount = CalculateCellCount()

        If _cellCount <> UBound(Width) Then
            Throw New Exception("Number of columns doesn't match the setting for column width.")
        End If

        For I = 1 To UBound(Width)
            totalWidth = totalWidth + Width(I)      ' Calculate the total
        Next

        If totalWidth <= 0 Then
            Throw New Exception("Can't set column width on table.")
        End If

        For Each row In _rows
            cols = 0
            For Each cell In row.cells
                cols = cols + cell.ColumnSpan
                cell.WidthInPercent = Math.Round(Width(cols) / totalWidth, 2)    ' Percent
            Next
        Next
    End Sub

    Private Function CalculateCellCount() As Integer
        Dim scellCnt As Integer, cellCnt As Integer, row As Row, cell As Cell
        For Each row In _rows
            cellCnt = 0
            For Each cell In row.cells
                cellCnt = cellCnt + cell.ColumnSpan
            Next
            If scellCnt <> 0 And scellCnt <> cellCnt Then
                Throw New Exception("Uneven number of cells with column span in the row collection.")
            End If
            scellCnt = cellCnt
        Next
        If cellCnt = 0 Then
            Throw New Exception("No columns/cells.")
        End If
        CalculateCellCount = cellCnt
    End Function

    Private Function line(ByVal x As Integer, ByVal y As Integer, ByVal x1 As Integer, ByVal y1 As Integer) As String
        Dim stream As String
        stream += x & " " & y & " m" + vbCr
        stream += x1 & " " & y1 & " l" + vbCr
        stream += "S" + vbCr
        line = stream
    End Function

    Public Property Border() As Borders
        Get
            Return _border
        End Get
        Set(ByVal myBorder As Borders)
            Select Case myBorder
                Case Borders.none
                    _border = myBorder
                Case Borders.thick
                    _border = myBorder
                Case Borders.thin
                    _border = myBorder
                Case Else
                    Throw New Exception("Invalid Border")
            End Select
        End Set
    End Property

    Public Sub AddRow(ByVal myRow As Row)
        _rows.Add(myRow)
    End Sub

    Public Overrides Function toString() As String Implements iDrawableObject.toString
        toString = "Table rows: " & _rows.Count
    End Function

    Public Property CellPaddingInPDFUnits() As Integer
        Get
            Return _cellPadding
        End Get
        Set(ByVal CellPad As Integer)
            If CellPad < 0 Then
                Throw New Exception("Invalid Cellpadding")
            End If
            _cellPadding = CellPad
        End Set
    End Property
End Class
