''''''''''''''''''''''''''''''''''''''''''''''''''
' Copyright www.acecomputing.biz
' Feel free to use and modify this code for private use.
' Not licensed for resale
'''''''''''''''''''''''''''''''''''''''''''''''''''
Public Class TextArea
    Implements iDrawableObject
    Public ReadOnly _copyrightAceComputing As String = "www.acecomputing.biz"

#If Debug Then
    Private eol = vbCrLf
#Else
    Private eol = vbLf
#End If

    Private _Texts As ArrayList      ' texts to be word wrapped
    Private _LineQ As Queue          ' word wrapped lines
    Private _Height As Integer
    Private _StartH As Integer = 72
    Private _StartV As Integer
    Private _widthPDFUnits As Integer

    Public Sub New()
        _Texts = New ArrayList()
    End Sub

    Public Sub New(ByVal text As String)
        Me.New()
        Me.AddText(text)
    End Sub

    Sub CalculateHeight(ByVal width As Integer)
        Dim MyText As TextObject
        Dim fontRef As String, key As String
        Dim sFontRef As String, found As Boolean
        Dim FontSize As Integer, sFontSize As Integer
        Dim i As Short
        Dim lineNo As Integer
        Dim linelen As Integer
        Dim textLine As TextObject
        Dim line As String          ' Text line
        Dim tmpline As String
        Dim vspace As Integer
        Dim ret As String
        Dim fontO As FontObj

        If width < 1 Then
            Throw New Exception("Invalid width for TextArea")
        End If
        _widthPDFUnits = width
        _LineQ = New Queue()
        ' Split the text up in lines
        For Each MyText In _Texts
            line = MyText.Text
            ' Escape PDF special characters ( and )
            line = ReplaceText(ReplaceText(line, "(", "\("), ")", "\)")
            line = Trim(line)
            FontSize = MyText.FontSize
            linelen = MyText.FontObj.HorizontalSpace * width / MyText.FontSize
            If Len(line) > linelen Then
                'word wrap
                Do While Len(line) > linelen
                    tmpline = line.Substring(0, linelen)
                    For i = Len(tmpline) To Len(tmpline) / 2 Step -1
                        If InStr("*&^%$#,. ;<=>[])}!""", Mid(tmpline, i, 1)) Then
                            ' find appropriate end of line
                            tmpline = tmpline.Substring(0, i)
                            Exit For
                        End If
                    Next
                    line = Mid(line, Len(tmpline) + 1)
                    textLine = New TextObject()
                    With textLine
                        .Text = tmpline
                        .FontObj = MyText.FontObj
                        .FontSize = MyText.FontSize
                    End With
                    _LineQ.Enqueue(textLine)
                Loop
                textLine = New TextObject()
                With textLine
                    .Text = line
                    .FontObj = MyText.FontObj
                    .FontSize = MyText.FontSize
                End With
                _LineQ.Enqueue(textLine)
            Else
                textLine = New TextObject()
                With textLine
                    .Text = line
                    .FontObj = MyText.FontObj
                    .FontSize = MyText.FontSize
                End With
                _LineQ.Enqueue(textLine)
            End If
        Next

        _Height = 0
        For Each MyText In _LineQ
            FontSize = MyText.FontSize
            _Height += 1.2 * FontSize
        Next
    End Sub

    Function Draw(ByRef StartV As Integer, ByVal width As Integer, ByRef FontAlias As Hashtable, _
        ByRef pagenum As Integer, ByVal TopStart As Integer) As PDFObject Implements iDrawableObject.draw
        Dim PDFO As PDFObject, MyText As TextObject, fontName As String, TempPdfo As PDFObject
        Dim fontRef As String, key As String
        Dim sFontRef As String, found As Boolean
        Dim FontSize As Integer, sFontSize As Integer
        Dim i As Short
        Dim lineNo As Integer
        Dim linelen As Integer
        Dim textLine As TextObject
        Dim line As String          ' Text line
        Dim tmpline As String
        Dim vspace As Integer
        Dim ret As String
        Dim fontO As FontObj
        Dim page As PageBreak
        Dim save As String

        Call CalculateHeight(width)

        PDFO = New PDFObject()
        page = New PageBreak()

        ' Process fonts
        For Each MyText In Me._Texts
            ret = MyText.Text
            ' Set if we have this font
            fontRef = getFontNumber(MyText.Font, MyText.FontStyle, FontAlias)

            If fontRef = "" Then
                'Add a new font 
                fontRef = Trim(Str(FontAlias.Count + 1))
                fontO = New FontObj()
                With fontO
                    .FontRef = fontRef
                    .Font = MyText.Font
                    .FontStyle = MyText.FontStyle
                End With
                FontAlias.Add(fontRef, fontO)
            End If

            MyText.FontObj.FontRef = fontRef
            found = False
            For Each key In PDFO._fonts
                If key.Equals(fontRef) Then found = True
            Next
            If found = False Then
                PDFO._fonts.Add(fontRef)
            End If
        Next

        ' Print the lines to the PDF document

        lineNo = -1
        ret = " BT" + vbCr                       ' Begin text object

        For Each MyText In _LineQ
            line = MyText.Text
            fontName = MyText.FontObj.FontName()
            fontRef = MyText.FontObj.FontRef
            FontSize = MyText.FontSize
            vspace = 1.2 * FontSize
            If (sFontRef <> fontRef) Or sFontSize <> FontSize Then
                ret += "/F" & fontRef & " " & FontSize & " Tf" & vbCr  ' Text and font
                ret += "1 0 0 1 " & _StartH & " " & StartV & " Tm" & vbCr   ' Set text matrix
                ret += vspace & " TL" & vbCr  ' Set text leading
                '               lineNo = lineNo + 1
            End If
            sFontRef = fontRef
            sFontSize = FontSize
            ret += "T* (" & line & eol & ") Tj" & vbCr
            StartV -= vspace
            If StartV < 100 Then
                ' Print footer
                page.StartPDFH = Me._StartH
                ret += "ET " + vbCrLf
                TempPdfo = page.Draw(StartV, width, FontAlias, pagenum, TopStart)
                ret += TempPdfo.getStream()
                PDFO.addStream(ret)
                PDFO.PageBreak = True
                ' Start new page
                save = ret
                ret = New String("")
                ret = "BT " + vbCrLf
                sFontRef = ""
                StartV = TopStart
            End If
        Next
        StartV -= vspace
        ret += " ET" + vbCr
        PDFO.addStream(ret)
        Draw = PDFO
    End Function

    Function GetCopy() As iDrawableObject Implements iDrawableObject.GetCopy
        Dim Text As TextArea, tobj As TextObject
        Text = New TextArea()
        For Each tobj In _Texts
            Text.AddText(tobj.Text, tobj.Font, tobj.FontSize, tobj.FontStyle)
        Next
        With Text
            .StartPDFH = Me._StartH
        End With
        Return Text
    End Function

    Property StartPDFH() As Integer Implements iDrawableObject.StartPDFH
        Get
            Return _StartH
        End Get
        Set(ByVal startH As Integer)
            If startH < 1 Then
                Throw New Exception("Invalid horizontal coordinate for cell.")
            End If
            _StartH = startH
        End Set
    End Property

    Public ReadOnly Property HeightInPDFunits() As Integer
        Get
            Return _Height
        End Get
    End Property

    Function getTexts() As ArrayList
        Return _Texts
    End Function

    <MonitoringDescription("Add word wrapped sentences.")> _
    Public Sub AddText(ByVal Text As String, Optional ByVal Font As Fonts = Fonts.Helvetica, _
        Optional ByVal fontSize As Integer = 10, Optional ByVal style As FontStyles = FontStyles.Regular)
        Dim myText As TextObject
        myText = New TextObject()
        With myText
            .Font = Font
            .FontSize = fontSize
            .Text = Text
            .FontStyle = style
        End With
        _Texts.Add(myText)
    End Sub

    Overrides Function toString() As String Implements iDrawableObject.toString
        Dim ret As String
        ret = "TextArea: "
        If _Texts.Count > 0 Then
            ret += _Texts.Item(0).text
        End If
        toString = ret
    End Function

    Property StartPDFV() As Integer
        Get
            Return _StartV
        End Get
        Set(ByVal startV As Integer)
            If startV < 1 Then
                Throw New Exception("Invalid vertical coordinate for cell.")
            End If
            _StartV = startV
        End Set
    End Property

    Private Function getFontNumber(ByVal Font As Fonts, ByVal fontstyle As FontStyle, ByRef fonts As Hashtable) As String
        Dim i As Integer, key As String, fName As String, fonto As FontObj
        For i = 1 To fonts.Count
            key = Trim(Str(i))
            fonto = fonts.Item(key)
            If fonto.Font = Font And fonto.FontStyle = fontstyle Then
                'If Font.equals(fonto) Then
                getFontNumber = fonto.FontRef
            End If
        Next
    End Function

    Public Function ReplaceText(ByRef Text_Renamed As String, ByRef TextToReplace As String, ByRef NewText As String) As String
        Dim mtext As String
        Dim SpacePos As Integer
        mtext = Text_Renamed
        SpacePos = InStr(mtext, TextToReplace)
        Do While SpacePos
            mtext = mtext.Substring(0, SpacePos - 1) & NewText & Mid(mtext, SpacePos + Len(TextToReplace))
            SpacePos = InStr(SpacePos + Len(NewText), mtext, TextToReplace)
        Loop
        ReplaceText = mtext
    End Function
End Class
